(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FWCompilation.thy --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Université Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FWCompilation.thy 9221 2011-08-17 13:52:01Z lukasbru $ *)

theory 
  LRFWCompilation
imports
  "../PacketFilter/PacketFilter"
begin

text{* This theory contains all the definitions used for policy
normalisation as described in~\cite{brucker.ea:icst:2010}.

The normalisation procedure transforms policies into semantically equivalent 
ones which are "easier" to test. It is
organized into nine phases. We impose the following two
restrictions on the input policies:
\begin{itemize}
\item Each policy must contain a $\mathtt{DenyAll}$ rule. If this
  restriction were to be lifted, the $\mathtt{insertDenies}$ phase
  would have to be adjusted accordingly.
\item For each pair of networks $n_1$ and $n_2$, the networks are
  either disjoint or equal. If this restriction were to be lifted, we
  would need some additional phases before the start of the
  normalisation procedure presented below. This rule would split
  single rules into several by splitting up the networks such that
  they are all pairwise disjoint or equal. Such a transformation is
  clearly semantics-preserving and the condition would hold after
  these phases.
\end{itemize}
As a result, the procedure generates a list of policies, in which:
\begin{itemize}
\item each element of the list contains a policy which completely
  specifies the blocking behavior between two networks, and
\item there are no shadowed rules.
\end{itemize}
This result is desirable since the test case generation for rules
between networks $A$ and $B$ is independent of the rules that specify
the behavior for traffic flowing between networks $C$ and $D$. Thus,
the different segments of the policy can be processed
individually. The normalization procedure does not aim to minimize the
number of rules. While it does remove unnecessary ones, it also adds
new ones, enabling a policy to be split into several independent
parts.
*}


text{* Policy transformations are
functions that map policies to policies.  We decided to represent
policy transformations as \emph{syntactic rules}; this choice paves
the way for expressing the entire normalisation process inside HOL by
functions manipulating abstract policy syntax. *}


subsection{* Basics *}
text{*
We define a very simple policy language:
*}

datatype ('\<alpha>,'\<beta>) Combinators = 
  DenyAll
| DenyAllFromTo '\<alpha> '\<alpha> 
| AllowPortFromTo '\<alpha> '\<alpha> '\<beta> 
| Conc "(('\<alpha>,'\<beta>) Combinators)" "(('\<alpha>,'\<beta>) Combinators)" (infixr "\<oplus>" 80)

text{* And define the semantic interpretation of it. For technical
reasons, we fix here the type to policies over IntegerPort
addresses. However, we could easily provide definitions for other
address types as well, using a generic consts for the type definition
and a primrec definition for each desired address model.  *}

subsection{* Auxiliary definitions and functions. *}

text{*This subsection defines several functions which are useful later
for the combinators, invariants, and proofs. *}

fun position :: "'\<alpha> \<Rightarrow> '\<alpha> list \<Rightarrow>  nat" where
  "position a [] = 0"
|"(position a (x#xs)) = (if a = x then 1 else (Suc (position a xs)))" 

fun srcNet where 
 "srcNet (DenyAllFromTo x y) = x"
|"srcNet (AllowPortFromTo x y p) = x"

fun destNet where 
 "destNet (DenyAllFromTo x y) = y"
|"destNet (AllowPortFromTo x y p) = y"

fun srcnets where
 "srcnets DenyAll = [] " 
|"srcnets (DenyAllFromTo x y) = [x] " 
|"srcnets (AllowPortFromTo x y p) = [x] " 
|"(srcnets (x \<oplus> y)) = (srcnets x)@(srcnets y)"

fun destnets where
 "destnets DenyAll = [] " 
|"destnets (DenyAllFromTo x y) = [y] " 
|"destnets (AllowPortFromTo x y p) = [y] "
|"(destnets (x \<oplus> y)) = (destnets x)@(destnets y)"

fun (sequential) net_list_aux where
 "net_list_aux [] = []"
|"net_list_aux (DenyAll#xs) = net_list_aux xs"
|"net_list_aux ((DenyAllFromTo x y)#xs) = x#y#(net_list_aux xs)"
|"net_list_aux ((AllowPortFromTo x y p)#xs) = x#y#(net_list_aux xs)"
|"net_list_aux ((x\<oplus>y)#xs) = (net_list_aux [x])@(net_list_aux [y])@(net_list_aux xs)"

fun net_list where "net_list p = remdups (net_list_aux p)"

definition bothNets where "bothNets x = (zip (srcnets x) (destnets x))"

fun (sequential) normBothNets where 
 "normBothNets ((a,b)#xs) = (if ((b,a) \<in> set xs) \<or> (a,b) \<in> set (xs) 
                            then (normBothNets xs)
                            else (a,b)#(normBothNets xs))"
|"normBothNets x = x" 

fun makeSets where 
 "makeSets ((a,b)#xs) = ({a,b}#(makeSets xs))"
|"makeSets [] = []"

fun bothNet where
 "bothNet DenyAll = {}"
|"bothNet (DenyAllFromTo a b) = {a,b}"
|"bothNet (AllowPortFromTo a b p) = {a,b}"

text{* $Nets\_List$ provides from a list of rules a list where the
entries are the appearing sets of source and destination network of each
rule. *}

definition Nets_List where "Nets_List x = makeSets (normBothNets (bothNets x))"

fun (sequential) first_srcNet where 
  "first_srcNet (x\<oplus>y) = first_srcNet x"
| "first_srcNet x = srcNet x"

fun (sequential) first_destNet where 
  "first_destNet (x\<oplus>y) = first_destNet x"
| "first_destNet x = destNet x"

fun (sequential) first_bothNet where
 "first_bothNet (x\<oplus>y) = first_bothNet x"
|"first_bothNet x = bothNet x"

fun (sequential) in_list where 
 "in_list DenyAll l = True"
|"in_list x l = (bothNet x \<in> set l)"

fun all_in_list where 
 "all_in_list [] l = True"
|"all_in_list (x#xs) l = (in_list x l \<and> all_in_list xs l)" 

fun (sequential) member where
 "member a (x\<oplus>xs) = ((member a x) \<or> (member a xs))"
|"member a x = (a = x)"

fun sdnets where
  "sdnets DenyAll = {}"
| "sdnets (DenyAllFromTo a b) = {(a,b)}"
| "sdnets (AllowPortFromTo a b c) = {(a,b)}"
| "sdnets (a \<oplus> b) = sdnets a \<union> sdnets b"

definition packet_Nets  where "packet_Nets x a b = ((src x \<sqsubset> a \<and> dest x \<sqsubset> b) \<or>
                                                    (src x \<sqsubset> b \<and> dest x \<sqsubset> a))"


definition subnetsOfAdr where "subnetsOfAdr a = {x. a \<sqsubset> x}"

definition fst_set where "fst_set s = {a. \<exists> b. (a,b) \<in> s}"

definition snd_set where "snd_set s = {a. \<exists> b. (b,a) \<in> s}"

fun memberP where
 "memberP r (x#xs) = (member r x \<or> memberP r xs)"
|"memberP r [] = False"

fun firstList where 
 "firstList (x#xs) = (first_bothNet x)"
|"firstList [] = {}"

subsection{* Invariants *}
text{* If there is a DenyAll, it is at the last position *}
fun wellformed_policy1:: "(('\<alpha>, '\<beta>) Combinators) list \<Rightarrow> bool" where 
  "wellformed_policy1 [] = True"
| "wellformed_policy1 (DenyAll#xs) = (xs=[])" 
| "wellformed_policy1 ((DenyAllFromTo a b)#xs) = (wellformed_policy1 xs)" 
| "wellformed_policy1 ((AllowPortFromTo a b p)#xs) = (wellformed_policy1 xs)" 
| "wellformed_policy1 ((a \<oplus> b)#xs) = (wellformed_policy1 xs)" 

text{* There is a DenyAll at the first position *}
fun wellformed_policy1_strong:: "(('\<alpha>, '\<beta>) Combinators) list \<Rightarrow> bool"
where 
 "wellformed_policy1_strong ((DenyAllFromTo a b)#xs) = (wellformed_policy1_strong xs)" 
| "wellformed_policy1_strong ((AllowPortFromTo a b p)#xs) = (wellformed_policy1_strong xs)" 
| "wellformed_policy1_strong ((a \<oplus> b)#xs) = (wellformed_policy1_strong xs)" 
| "wellformed_policy1_strong (DenyAll#xs) = (xs = [])" 
| "wellformed_policy1_strong [] = (False) "


text{* All two networks are either disjoint or equal. *}
definition netsDistinct where "netsDistinct a b = (\<not> (\<exists> x. x \<sqsubset> a \<and> x \<sqsubset> b))"

definition twoNetsDistinct where
  "twoNetsDistinct a b c d = (netsDistinct a c \<or> netsDistinct b d)"

definition allNetsDistinct where 
  "allNetsDistinct p = (\<forall> a b. (a \<noteq> b \<and> a \<in> set (net_list p) \<and>
                        b \<in> set (net_list p)) \<longrightarrow> netsDistinct a b)"

definition disjSD_2 where 
 "disjSD_2 x y = (\<forall> a b c d. ((a,b)\<in>sdnets x \<and>  (c,d) \<in>sdnets y \<longrightarrow>
      (twoNetsDistinct a b c d \<and> twoNetsDistinct a b d c)))"

text{* The policy is given as a list of single rules. *}
fun singleCombinators where 
"singleCombinators [] = True"
|"singleCombinators ((x\<oplus>y)#xs) = False"
|"singleCombinators (x#xs) = singleCombinators xs"

definition onlyTwoNets where
 "onlyTwoNets x = (x=DenyAll \<or> (\<exists> a b. (sdnets x = {(a,b)})) \<or> (\<exists> a b. sdnets x = {(a,b),(b,a)}))" 

text{* Each entry of the list contains rules between two networks only. *}
fun OnlyTwoNets where 
"OnlyTwoNets (x#xs) = (onlyTwoNets x \<and> OnlyTwoNets xs)"
|"OnlyTwoNets [] = True"

fun noDenyAll where
 "noDenyAll (x#xs) = ((\<not> member DenyAll x) \<and> noDenyAll xs)"
|"noDenyAll [] = True"

definition noDenyAll1 where
 "noDenyAll1 x = (noDenyAll x \<or> (last x = DenyAll \<and> noDenyAll (butlast x)))"

fun separated where
  "separated (x#xs) = ((\<forall> s. s \<in> set xs \<longrightarrow> disjSD_2 x s) \<and> separated xs)"
| "separated [] = True"

fun NetsCollected where 
  "NetsCollected (x#xs) = (((first_bothNet x \<noteq> firstList xs) \<longrightarrow>
          (\<forall>a\<in>set xs. first_bothNet x \<noteq> first_bothNet a)) \<and> NetsCollected (xs))"
| "NetsCollected [] = True"

fun NetsCollected2 where
 "NetsCollected2 (x#xs) = (xs = [] \<or> (first_bothNet x \<noteq> firstList xs \<and>
                           NetsCollected2 xs))"
|"NetsCollected2 [] = True"

subsection{* Transformations *}

text {*
The following two functions transform a policy into a list of single rules and vice-versa.
*}
fun policy2list::"('\<alpha>, '\<beta>) Combinators \<Rightarrow>
                 (('\<alpha>, '\<beta>) Combinators) list" where
 "policy2list (x \<oplus> y) = (concat [(policy2list x),(policy2list y)])"
|"policy2list x = [x]"

fun list2policy::"(('\<alpha>, '\<beta>) Combinators) list \<Rightarrow>
                  (('\<alpha>, '\<beta>) Combinators)" where
 "list2policy (x#[]) = x"
| "list2policy (x#y) = x \<oplus> (list2policy y)"

text{* Remove all the rules appearing after a DenyAll. There are two alternative versions. *}

fun removeShadowRules1 where
  "removeShadowRules1 (x#xs) = (if (x=DenyAll) then [x] 
                                else (x#(removeShadowRules1 xs)))"
| "removeShadowRules1 [] = []"

fun is_allow where 
  "is_allow (AllowPortFromTo a b p) (x, y) = (a = x \<and> b = y)"
| "is_allow x y = False"

text{* Remove all the rules which allow a port, but are shadowed by a deny between these subnets,
       *}

primrec
  removeAllP :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list" where
    "removeAllP P [] = []"
  | "removeAllP P (y # xs) = (if P y then removeAllP P xs else y # removeAllP P xs)"

subsubsection {* @{text removeAllP} *}
(*
lemma removeAllP_filter_not_eq:
  "removeAllP x = filter (\<lambda>y. x \<noteq> y)"
proof
  fix xs
  show "removeAllP x xs = filter (\<lambda>y. x \<noteq> y) xs"
    by (induct xs) auto
qed

lemma removeAllP_append[simp]:
  "removeAllP x (xs @ ys) = removeAllP x xs @ removeAllP x ys"
by (induct xs) auto

lemma set_removeAllP[simp]: "set(removeAllP x xs) = set xs - {x}"
by (induct xs) auto

lemma removeAllP_id[simp]: "x \<notin> set xs \<Longrightarrow> removeAllP x xs = xs"
by (induct xs) auto

(* Needs count:: 'a \<Rightarrow> a' list \<Rightarrow> nat
lemma length_removeAllP:
  "length(removeAllP x xs) = length xs - count x xs"
*)

lemma removeAllP_filter_not[simp]:
  "\<not> P x \<Longrightarrow> removeAllP x (filter P xs) = filter P xs"
by(induct xs) auto

lemma distinct_removeAllP:
  "distinct xs \<Longrightarrow> distinct (removeAllP x xs)"
  by (simp add: removeAllP_filter_not_eq)

lemma distinct_remove1_removeAllP:
  "distinct xs ==> remove1 x xs = removeAllP x xs"
by (induct xs) simp_all

lemma map_removeAllP_inj_on: "inj_on f (insert x (set xs)) \<Longrightarrow>
  map f (removeAllP x xs) = removeAllP (f x) (map f xs)"
by (induct xs) (simp_all add:inj_on_def)

lemma map_removeAllP_inj: "inj f \<Longrightarrow>
  map f (removeAllP x xs) = removeAllP (f x) (map f xs)"
by(metis map_removeAllP_inj_on subset_inj_on subset_UNIV)

*)



fun removeShadowRules2::  "(('\<alpha>, '\<beta>) Combinators) list \<Rightarrow>
                          (('\<alpha>, '\<beta>) Combinators) list"
where
  "removeShadowRules2 ((DenyAllFromTo x y)#z) = 
  ((DenyAllFromTo x y)#(removeShadowRules2 (filter (\<lambda> p. \<not> is_allow p (x,y)) z)))"
| "removeShadowRules2 (x#y) = x#(removeShadowRules2 y)"
| "removeShadowRules2 [] = []"

text{* Sorting a pocliy.  We first need to define an ordering on
rules.  This ordering depends on the $Nets\_List$ of a policy.  *}

fun smaller :: "('\<alpha>, '\<beta>) Combinators \<Rightarrow> 
                ('\<alpha>, '\<beta>) Combinators \<Rightarrow> 
                (('\<alpha>) set) list \<Rightarrow> bool" 
where
  "smaller DenyAll x l = (x = DenyAll)" 
| "smaller x DenyAll l = True"
| "smaller x y l = 
 ((x = y) \<or> (if (bothNet x) = (bothNet y) then 
                (case x of (DenyAllFromTo a b) \<Rightarrow> (y = DenyAllFromTo b a)  
                           | _ \<Rightarrow>  True)
            else
                (position (bothNet x) l <= position (bothNet y) l)))"


text{*
We use insertion sort for sorting a policy.
*}

fun insort where
  "insort a [] l = [a]"
| "insort a (x#xs) l = (if (smaller a x l) then a#x#xs else x#(insort a xs l))"

fun sort where
  "sort [] l = []"
| "sort (x#xs) l = insort x (sort xs l) l"

fun sorted  where
"sorted [] l \<longleftrightarrow> True" |
"sorted [x] l \<longleftrightarrow> True" |
"sorted (x#y#zs) l \<longleftrightarrow> smaller x y l \<and> sorted (y#zs) l"

text{*
@{text separate} works on a sorted policy: it joins the rules which talk about the traffic between the same two networks.
*}

fun separate where
 "separate (DenyAll#x) = DenyAll#(separate x)"
|"separate (x#DenyAll#y) = x#DenyAll#(separate y)"
| "separate (x#y#z) = (if (first_bothNet x = first_bothNet y) 
                   then (separate ((x\<oplus>y)#z))
                   else (x#(separate(y#z))))"
|"separate x = x"                

text {*
Insert the DenyAllFromTo rules, such that traffic between two networks can be tested individually
*}

fun  insertDenies where
 "insertDenies (DenyAll#xs) = (DenyAll#(insertDenies xs))"
|"insertDenies (x#xs) = (x \<oplus> (DenyAllFromTo (first_srcNet x) (first_destNet x)) \<oplus> 
                            (DenyAllFromTo (first_destNet x) (first_srcNet x)))#
                                (insertDenies xs)"
| "insertDenies [] = []"


fun  insertDenies1 where
 "insertDenies1 [DenyAll] = [DenyAll]"
|"insertDenies1 (x#xs) = (case x of DenyAll \<Rightarrow> (DenyAll#(insertDenies1 xs))
                    | _  \<Rightarrow> (x \<oplus> (DenyAllFromTo (first_srcNet x) (first_destNet x)) \<oplus> 
                            (DenyAllFromTo (first_destNet x) (first_srcNet x)))#
                                (insertDenies1 xs))"
| "insertDenies1 [] = []"

thm insertDenies.simps insertDenies1.simps

text{* Remove duplicate rules. This is especially necessary as
insertDenies might have inserted duplicate rules.

The second function is supposed to work on a list of policies. Only
rules which are duplicated within the same policy are removed.  *}


function remdupsP where
 "remdupsP (x#xs) = (remdupsP (removeAll x xs))"
|"remdupsP [] = []"
apply simp_all
by (metis list.exhaust)


fun  remdups2 :: "'a list \<Rightarrow> 'a list" where
  "remdups2 (x # xs) = x#(remdups2 (filter (\<lambda>y. (y\<noteq>x)) xs))"
| "remdups2 [] = []"

fun removeDuplicates where
  "removeDuplicates (x\<oplus>xs) = (if member x xs then (removeDuplicates xs)
                              else x\<oplus>(removeDuplicates xs))"
| "removeDuplicates x = x"

fun removeAllDuplicates where 
 "removeAllDuplicates (x#xs) = ((removeDuplicates (x))#(removeAllDuplicates xs))"
|"removeAllDuplicates x = x"



text {* Insert a DenyAll at the end of a policy. *}
fun insertDeny where 
 "insertDeny [] = [DenyAll]"
|"insertDeny (xs) = (if last xs = DenyAll then xs else (xs@[DenyAll]))"


definition "sort' p l = sort l p"


declare dom_eq_empty_conv [simp del]





fun list2policyR::"(('\<alpha>, '\<beta>) Combinators) list \<Rightarrow>
                   (('\<alpha>, '\<beta>) Combinators)" where
 "list2policyR (x#[]) = x"
| "list2policyR (x#y) = (list2policyR y) \<oplus> x"






text{* We provide the definitions for two address representations. *}


subsection{* InterPort *}


fun C :: "(IntegerPort net, port) Combinators \<Rightarrow> (IntegerPort,DummyContent) packet \<mapsto> unit"
where
" C DenyAll = deny_all" 
|"C (DenyAllFromTo x y) = deny_all_from_to x y"
|"C (AllowPortFromTo x y p) = allow_from_to_port p x y"
|"C  (x \<oplus> y) =  C x \<Oplus> C y"

fun notMTpolicy::"((IntegerPort net, port) Combinators) list \<Rightarrow> bool"  where
 "notMTpolicy (x#xs) = (if (dom (C x) = {}) then (notMTpolicy xs) else True)"
|"notMTpolicy [] = False"

fun  matching_rule where
"matching_rule a (x#xs) = (if a \<in> dom (C x) then (Some x)
                               else (matching_rule a xs))"
|"matching_rule a [] = None"


fun noneMT where 
 "noneMT (x#xs) = (dom (C x) \<noteq> {} \<and>  (noneMT xs))"
|"noneMT [] = True" 

text{* All rules appearing at the right of a DenyAllFromTo, have
disjunct domains from it (except DenyAll) *}

fun (sequential) wellformed_policy2 where
  "wellformed_policy2 [] = True"
| "wellformed_policy2 ((DenyAllFromTo a b)#xs) = ((\<forall> c. c \<in> set xs \<and> c \<noteq> DenyAll \<longrightarrow>
                 Map.dom (C (DenyAllFromTo a b)) \<inter> Map.dom (C c) = {}) \<and> wellformed_policy2 xs)"
| "wellformed_policy2 (x#xs) = wellformed_policy2 xs"



text{* An allow rule is disjunct with all rules appearing at the left
of it. This invariant is not necessary as it is a consequence from
others, but facilitates some proofs. *}

fun (sequential) wellformed_policy3 where
  "wellformed_policy3 [] = True"
| "wellformed_policy3 (DenyAll#xs) = wellformed_policy3 xs"
| "wellformed_policy3 (x#xs) = ((\<forall> c a b p. c = AllowPortFromTo a b p \<and> c \<in> set xs \<longrightarrow>
                 Map.dom (C x) \<inter> Map.dom (C c) = {}) \<and> wellformed_policy3 xs)"



fun removeShadowRules3::"((IntegerPort net, port) Combinators) list \<Rightarrow> 
                         ((IntegerPort net, port) Combinators) list"  where
 "removeShadowRules3 (x#xs) = (if (dom (C x)= {}) then (removeShadowRules3 xs)
                               else (x#(removeShadowRules3 xs)))"
|"removeShadowRules3 [] = []"


definition 
  "normalize' p  = (removeAllDuplicates o insertDenies o separate o
                   (sort' (Nets_List p))  o removeShadowRules2 o remdups2 o
                   removeShadowRules3 o insertDeny o removeShadowRules1 o
                   policy2list) p"


definition normalize :: 
  "(IntegerPort net, port) Combinators \<Rightarrow> 
   (IntegerPort net, port) Combinators list" 
where
   "normalize p = (removeAllDuplicates (insertDenies (separate (sort
                   (removeShadowRules2 (remdups2 (removeShadowRules3 (insertDeny
                  (removeShadowRules1 (policy2list p)))))) ((Nets_List p))))))"

definition
   "normalize_manual_order p l = removeAllDuplicates (insertDenies (separate
    (sort (removeShadowRules2 (remdups2 (removeShadowRules3 (insertDeny
    (removeShadowRules1 (policy2list p)))))) ((l)))))"

text{* Of course, normalize is equal to normalize', the latter looks nicer though. *}
lemma "normalize = normalize'"
by (rule ext, simp add: normalize_def normalize'_def sort'_def)

text{* The following definition helps in creating the test specification for the individual parts of a normalized policy. OBSOLETE NOW.*}


definition makeFUT where 
   "makeFUT FUT p x n = 
     (packet_Nets x (fst (normBothNets (bothNets p)!n))  
                     (snd(normBothNets (bothNets p)!n)) \<longrightarrow> 
    FUT x = C ((normalize p)!Suc n) x)"


declare C.simps [simp del]


subsection{* TCP\_UDP\_IntegerPort *}



fun Cp :: "(IntegerPort_TCPUDP net, protocol \<times> port) Combinators \<Rightarrow> 
          (IntegerPort_TCPUDP,DummyContent) packet \<mapsto> unit"
where
 " Cp DenyAll = deny_all"  
|"Cp (DenyAllFromTo x y) = deny_all_from_to x y"
|"Cp (AllowPortFromTo x y p) = allow_from_to_port_prot (fst p) (snd p) x y"
|"Cp  (x \<oplus> y) =  Cp x \<Oplus> Cp y"

fun notMTpolicyPr::"((IntegerPort_TCPUDP net, protocol \<times> port) Combinators) list \<Rightarrow> bool"  where
 "notMTpolicyPr (x#xs) = (if (dom (Cp x) = {}) then (notMTpolicyPr xs) else True)"
|"notMTpolicyPr [] = False"

text{* Provides the first matching rule of a policy given as a list of rules. *}


fun matching_rulePr where
"matching_rulePr a (x#xs) = (if a \<in> dom (Cp x) then (Some x)
                               else (matching_rulePr a xs))"
|"matching_rulePr a [] = None"


fun noneMTPr where 
 "noneMTPr (x#xs) = (dom (Cp x) \<noteq> {} \<and>  (noneMTPr xs))"
|"noneMTPr [] = True" 

text{* All rules appearing at the left of a DenyAllFromTo, have disjunct domains from it (except DenyAll) *}

fun (sequential) wellformed_policy2Pr where
  "wellformed_policy2Pr [] = True"
| "wellformed_policy2Pr ((DenyAllFromTo a b)#xs) = ((\<forall> c. c \<in> set xs \<and> c \<noteq> DenyAll \<longrightarrow>
                 Map.dom (Cp (DenyAllFromTo a b)) \<inter> Map.dom (Cp c) = {}) \<and> wellformed_policy2Pr xs)"
| "wellformed_policy2Pr (x#xs) = wellformed_policy2Pr xs"

text{* An allow rule is disjunct with all rules appearing at the right of it. This invariant is not necessary as it is a consequence from others, but facilitates some proofs. *}


fun (sequential) wellformed_policy3Pr where
  "wellformed_policy3Pr [] = True"
| "wellformed_policy3Pr (DenyAll#xs) = wellformed_policy3Pr xs"
| "wellformed_policy3Pr (x#xs) = ((\<forall> c a b p. c = AllowPortFromTo a b p \<and> c \<in> set xs \<longrightarrow>
                 Map.dom (Cp x) \<inter> Map.dom (Cp c) = {}) \<and> wellformed_policy3Pr xs)"



fun removeShadowRules3Pr  where
 "removeShadowRules3Pr (x#xs) = (if (dom (Cp x)= {}) then (removeShadowRules3Pr xs)
                               else (x#(removeShadowRules3Pr xs)))"
|"removeShadowRules3Pr [] = []"


definition 
 normalizePr' :: "(IntegerPort_TCPUDP net, protocol \<times> port) Combinators
  \<Rightarrow> (IntegerPort_TCPUDP net, protocol \<times> port) Combinators list" where 
 "normalizePr' p = (removeAllDuplicates o insertDenies o separate o
                   (sort' (Nets_List p))  o removeShadowRules2 o remdups2 o
                   removeShadowRules3Pr o insertDeny o removeShadowRules1 o
                   policy2list) p"


definition normalizePr :: 
"(IntegerPort_TCPUDP net, protocol \<times> port) Combinators
  \<Rightarrow> (IntegerPort_TCPUDP net, protocol \<times> port) Combinators list" where
   "normalizePr p = (removeAllDuplicates (insertDenies (separate (sort
                   (removeShadowRules2 (remdups2 (removeShadowRules3Pr (insertDeny
                  (removeShadowRules1 (policy2list p)))))) ((Nets_List p))))))"

definition
   "normalize_manual_orderPr p l = removeAllDuplicates (insertDenies (separate
    (sort (removeShadowRules2 (remdups2 (removeShadowRules3Pr (insertDeny
    (removeShadowRules1 (policy2list p)))))) l)))"


text{* Of course, normalize is equal to normalize', the latter looks nicer though. *}
lemma "normalizePr = normalizePr'"
by (rule ext, simp add: normalizePr_def normalizePr'_def sort'_def)

text{* The following definition helps in creating the test specification for the individual parts of a normalized policy. *}
definition makeFUTPr where 
   "makeFUTPr FUT p x n = 
     (packet_Nets x (fst (normBothNets (bothNets p)!n))  
                     (snd(normBothNets (bothNets p)!n)) \<longrightarrow> 
    FUT x = Cp ((normalizePr p)!Suc n) x)"


declare Cp.simps [simp del]



lemmas PLemmas = C.simps Cp.simps dom_def aux aux2
                 ProtocolPortCombinators.ProtocolCombinators src_def
                 dest_def in_subnet_def IntegerPort.src_port_def
                 IntegerPort.dest_port_def
                 IntegerPort_TCPUDP.src_port_def
                 IntegerPort_TCPUDP.dest_port_def
                 IntegerPort_TCPUDP.dest_protocol_def



lemma aux: "\<lbrakk>x \<noteq> a; y\<noteq>b; (x \<noteq> y \<and> x \<noteq> b) \<or>  (a \<noteq> b \<and> a \<noteq> y)\<rbrakk> \<Longrightarrow> {x,a} \<noteq> {y,b}"
  apply auto
done


lemma aux2: "{a,b} = {b,a}"
  apply auto
done



end
